@@ -27,7 +27,7 @@ from multiprocessing import get_context, Queue
 import numpy as np
 import psutil
 from time import sleep
-from typing import Callable, Dict, Iterable, List, Tuple, Union
+from typing import Callable, Dict, Generator, Iterable, List, Tuple, Union
 
 from sketchnu.countmin import CountMin, CountMinLinear
 from sketchnu.heavyhitters import HeavyHitters
@@ -132,10 +132,9 @@ def _log_worker(log_queue: Queue):
 def _worker(
     worker_id: int,
     sketch: Tuple,
-    process_q_item: Callable[..., Iterable[bytes]],
+    process_q_item: Callable[..., int],
     in_queue: Queue,
     log_queue: Queue,
-    ngram,
     **kwargs,
 ):
     """
@@ -154,16 +153,15 @@ def _worker(
         "hll" | "cms", sketch_args are dictionary of arguments to give to HyperLogLog
         or CountMin, and shm_name is the name of the shared memory block to attach the
         instantiated sketch to.
-    process_q_item : Callable[..., Iterable[bytes]]
-        Function that turns an item from the queue into an Iterator of
-        elements (bytes) to be added to the sketch(s).
+    process_q_item : Callable[..., int]
+        User defined function whose arguments are (q_item, \*sketches, \*\*kwargs) that
+        takes the q_item, adds elements to the sketch(es) and returns the number of
+        records that were processed. \*sketches must be in alphabetic order since that is
+        how they will be passed by `parallel_add`.
     in_queue : Queue
         Queue containing items to be processed by the workers
     log_queue : Queue
         Queue to send logging statements to
-    ngram : int
-        If 0 or None, then calls sketch.add(key). Otherwise calls
-        sketch.add_ngram(key, ngram)
     \*\*kwargs :
         Keyword arguments passed to process_q_item(q_item, \*\*kwargs)
 
@@ -186,13 +184,18 @@ def _worker(
         q_item = in_queue.get()
         # Process a queue_item
         if q_item is not None:
-            for keys in process_q_item(q_item, **kwargs):
-                for local_sketch in local_sketches:
-                    if ngram:
-                        local_sketch.update_ngram(keys, ngram)
-                    else:
-                        local_sketch.update(keys)
-                n_records += 1
+            try:
+                n_recs = process_q_item(q_item, *local_sketches, **kwargs)
+            except Exception as exc:
+                n_recs = 0
+                msg = f"WORKER {worker_id:02} threw exception on {q_item}: {exc}"
+                log_queue.put(
+                    {
+                        "level": "ERROR",
+                        "text": msg,
+                    }
+                )
+            n_records += n_recs
 
             end = datetime.now()
             speed = n_records / (end - start).total_seconds()
@@ -230,9 +233,8 @@ def _worker(
 
 def parallel_add(
     items: Iterable,
-    process_q_item: Callable[..., Iterable[Union[Iterable[bytes], Dict[bytes, int]]]],
+    process_q_item: Callable[..., int],
     n_workers: int = None,
-    ngram: int = None,
     cms_args: Dict = None,
     hh_args: Dict = None,
     hll_args: Dict = None,
@@ -240,40 +242,44 @@ def parallel_add(
 ):
     """
     Places `items` onto a queue to be processed by `n_workers` independent spawned
-    processes. The function `process_q_item` takes a single item and returns an
-    Iterable of records with each record being an either an Iterable of keys (bytes)
-    or a dictionary with key=byte, value=number to times to add that key to be added
-    to the sketch(s). The \*\*kwargs are passed along to `process_q_item` to allow for
-    any needed additional parameters. Once all items have been processed, the
-    `n_workers` sketch(s) are merged with the final result(s) being returned.
+    processes.
 
-    **Note**
-    If your data has duplicate keys within a q_item, you will likely see better
-    performance if `process_q_item` does ```yield Counter(keys)``` instead of just
-    ```yield keys```
+    The user defined function, `process_q_item`, takes the arguments
+    (q_item, \*sketches, \*\*kwargs). This function is given a single `item` from the
+    queue and then should add elements to the sketch(es) as desired. The function
+    should return the number of records processed. If `process_q_item` will add
+    elements to multiple sketches, then they must be listed in alphabetic order since
+    that is how `parallel_add` will pass them to `process_q_item`.
+
+    The \*\*kwargs are passed along to `process_q_item` to allow for any needed
+    additional parameters.
+
+    Once all `items` have been processed, the `n_workers` sketch(s) are merged with the
+    final result(s) returned.
 
     You must provide at least `cms_args` | `hh_args` | `hll_args`. If you provide more
     than one, then the requested sketches will be processed at the same time while
     going over the data just once.
 
+    **Note:** If your data has duplicate keys within a `item`, you will likely see
+    better performance if `process_q_item` does ```yield Counter(keys), n_records```
+    instead of just ```yield keys, n_records```
+
+
     Parameters
     ----------
     items : Iterable
         A generator or list of items that will be placed onto a queue and then worked
         by one of the workers in a separate spawned process.
-    process_q_item : Callable[...,Iterable[Union[Iterable[bytes],Dict[bytes,int]]]]
-        Function that turns an item from the queue into an Iterable of records with
-        each record being an Iterable of keys (bytes) to be added to the sketch(s)
-        or a Dict[key=bytes, value=int] to add each key multiple times to the
-        sketch(s). Using a Dict is often faster if data has multiple entries for a
-        given key.
+    process_q_item : Callable[..., int]
+        User defined function whose arguments are (q_item, \*sketches, \*\*kwargs) that
+        takes the q_item, adds elements to the sketch(es) and returns the number of
+        records that were processed. \*sketches must be in alphabetic order since that is
+        how they will be passed by `parallel_add`.
     n_workers : int, optional
         Number of workers to use. Each will update their own sketches which will then
         get merged together to achieve the final sketch(s). If None (default), then set
         to psutil.cpu_count(logical=False)
-    ngram : int, optional
-        If nonzero, then break each key into ngrams of this size before added to the
-        sketch(s)
     cms_args : Dict, optional
         Dictionary containing arguments to instantiate a CountMin.  If None (default)
         then don't create a sketch of this type.
@@ -284,7 +290,7 @@ def parallel_add(
         Dictionary containing arguments to instantiate a HyperLogLog. If None (default)
         then don't create a sketch of this type.
     \*\*kwargs :
-        Keyword arguments that get passed to `process_q_item` function
+        Keyword arguments that get passed to `process_q_item` generator function
 
     Returns
     -------
@@ -295,9 +301,6 @@ def parallel_add(
         raise ValueError("You forgot to provide any sketch arguments")
     if n_workers is None:
         n_workers = max(1, psutil.cpu_count(logical=False))
-    if ngram is not None:
-        if ngram <= 0:
-            raise ValueError(f"ngram is {ngram}. Must be positive")
 
     ctx = get_context("spawn")
     queue = ctx.Queue(3 * n_workers)
@@ -330,19 +333,19 @@ def parallel_add(
             cms_array.append(CountMin(**cms_args, shared_memory=True))
             # Send sketch type, its args, and shared memory block name
             sketch.append(("cms", cms_array[i].args, cms_array[i].shm.name))
-        if hll_args:
-            hll_array.append(HyperLogLog(**hll_args, shared_memory=True))
-            # Send sketch type, its args, and shared memory block name
-            sketch.append(("hll", hll_array[i].args, hll_array[i].shm.name))
         if hh_args:
             hh_array.append(HeavyHitters(**hh_args, shared_memory=True))
             # Send sketch type, its args, and shared memory block name
             sketch.append(("hh", hh_array[i].args, hh_array[i].shm.name))
+        if hll_args:
+            hll_array.append(HyperLogLog(**hll_args, shared_memory=True))
+            # Send sketch type, its args, and shared memory block name
+            sketch.append(("hll", hll_array[i].args, hll_array[i].shm.name))
         sketch = tuple(sketch)
         workers.append(
             ctx.Process(
                 target=_worker,
-                args=(i, sketch, process_q_item, queue, log_queue, ngram),
+                args=(i, sketch, process_q_item, queue, log_queue),
                 kwargs=kwargs,
             )
         )
